/*
 * Author: Hiroyuki Chishiro
 * License: 2-Clause BSD
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>
#include <arpa/inet.h>

#define PORT 65432
#define MAX_SEQ_ID 10
#define BUFSIZE 128
#define TIMEOUT_SEC 3
#define N 4
#define ACK "ACK"

int main(void)
{
  int sock;
  struct sockaddr_in client;
  struct timeval timeout;
  fd_set set;
  int ret;
  char buf[BUFSIZE], buf2[BUFSIZE];
  int i, j, n;
  int window_ack_index;

  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    perror("socket");
    exit(1);
  }

  memset(&client, 0, sizeof(client));
  client.sin_family = AF_INET;
  client.sin_port = htons(PORT);
  client.sin_addr.s_addr = inet_addr("127.0.0.1");

  if (connect(sock, (struct sockaddr *) &client, sizeof(client)) == -1) {
    perror("connect");
    goto end;
  }

  for (i = 0; i <= MAX_SEQ_ID; i += window_ack_index) {
    window_ack_index = 0;

    for (j = i; j < i + N && j <= MAX_SEQ_ID; j++) {
      memset(buf, 0, sizeof(buf));
      sprintf(buf, "%d", j);
      printf("Message to server: %s\n", buf);

      if (write(sock, buf, sizeof(buf)) == -1) {
        perror("write");
        goto end;
      }

      usleep(1000);
    }

    for (j = i; j < i + N && j <= MAX_SEQ_ID; j++) {
      FD_ZERO(&set);
      FD_SET(sock, &set);
      timeout.tv_sec = TIMEOUT_SEC;
      timeout.tv_usec = 0;

      if ((ret = select(sock + 1, &set, NULL, NULL, &timeout)) == -1) {
        perror("select");
        goto end;
      } else if (ret == 0) {
        printf("Going back from %d: timeout\n", i);
        break;
      } else {
        if ((n = read(sock, buf, sizeof(buf))) == -1) {
          perror("read");
          goto end;
        }

        buf[n] = '\0';
        printf("Message from server: %s\n", buf);
        sprintf(buf2, "%s%d", ACK, i + window_ack_index);

        if (strcmp(buf, buf2) == 0) {
          window_ack_index++;
        } else {
          break;
        }
      }
    }
  }

end:

  if (close(sock) == -1) {
    perror("close");
    exit(2);
  }

  return 0;
}
