/*
 * Author: Hiroyuki Chishiro
 * License: 2-Clause BSD
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <fcntl.h>

#define PORT 65432
#define MAX_SEQ_ID 10
#define BACKLOG 1024
#define BUFSIZE 128
#define ACK "ACK"
#define DISCARD_ID 5

int main(void)
{
  int s_sock, c_sock;
  struct sockaddr_in server, other;
  socklen_t other_size;
  int i, j, n;
  char buf[BUFSIZE];
  bool flag = true;

  if ((s_sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    perror("socket");
    exit(1);
  }

  memset(&server, 0, sizeof(server));
  memset(&other, 0, sizeof(other));
  server.sin_family = AF_INET;
  server.sin_port = htons(PORT);
  server.sin_addr.s_addr = INADDR_ANY;

  if (bind(s_sock, (struct sockaddr *) &server, sizeof(server)) == -1) {
    perror("bind");
    goto end;
  }

  if (listen(s_sock, BACKLOG) == -1) {
    perror("listen");
    goto end;
  }

  other_size = sizeof(other);

  if ((c_sock = accept(s_sock, (struct sockaddr *) &other, &other_size)) == -1) {
    perror("accept");
    goto end;
  }

  for (i = 0; i <= MAX_SEQ_ID; i++) {
    memset(buf, 0, sizeof(buf));

    if (i == DISCARD_ID && flag) {
      printf("i = %d: [Simulation loss]\n", i);
      flag = false;

      if (read(c_sock, buf, sizeof(buf)) == -1) {
        perror("read");
        goto end2;
      }
    }

    if ((n = read(c_sock, buf, sizeof(buf))) == -1) {
      perror("read");
      goto end2;
    }

    buf[n] = '\0';

    printf("Message from client: %s\n", buf);

    j = atoi(buf);

    if (j != i) {
      printf("Discarded as out of order\n");
      i--;
    } else {
      memset(buf, 0, sizeof(buf));
      sprintf(buf, "%s%d", ACK, i);
      printf("Message to client: %s\n", buf);

      if (write(c_sock, buf, sizeof(buf)) == -1) {
        perror("write");
        goto end2;
      }
    }
  }

end2:

  if (close(c_sock) == -1) {
    perror("close");
    exit(2);
  }

end:

  if (close(s_sock) == -1) {
    perror("close");
    exit(3);
  }

  return 0;
}
